#include <xen/arch-x86_32.h>

ES		    = 0x20
ORIG_EAX	= 0x24
EIP		    = 0x28
CS		    = 0x2C

// See notes in types.h regarding the exception stack frame, stack pointer,
// and privilege level change.

#define EXPORT_SYMBOL(x) .globl x; x:

#define SAVE_ALL \
	cld; \
	pushl %es; \
	pushl %ds; \
	pushl %eax; \
	pushl %ebp; \
	pushl %edi; \
	pushl %esi; \
	pushl %edx; \
	pushl %ecx; \
	pushl %ebx; \
	movl $(FLAT_KERNEL_DS),%edx; \
	movl %edx,%ds; \
	movl %edx,%es;

#define RESTORE_ALL	\
	popl %ebx;	\
	popl %ecx;	\
	popl %edx;	\
	popl %esi;	\
	popl %edi;	\
	popl %ebp;	\
	popl %eax;	\
	popl %ds;	\
	popl %es;	\
	addl $4,%esp;		

// Exception stack setup for IA32
// Upon entry, an exception frame will be on top of the stack as
// defined in System Programming I, Figure 6-4.  Note that if no privilege
// level change occurs when handling the interrupt, the old stack pointer is
// not pushed onto the stack & is not updated: in this case, the processor
// assumes that the existing stack will be used while handling the interrupt.
.macro stackenter
    mov %ss, orig_far_sp + 4
    mov %esp, orig_far_sp       // Save the original stack pointer
    lss exp_far_sp, %esp        // Load the exception stack pointer into SS:ESP
    sub $(6*4), %esp            // Make room on the stack for an exception frame
    SAVE_ALL                    // Save all registers
    les orig_far_sp, %esi       // Load the saved SP into ES:ESI
    movl %es:(%esi), %eax       
    movl %eax, 10*4(%esp)       // Copy the saved EIP into the new exp. frame
    movl %es:2*4(%esi), %eax    
    movl %eax, 12*4(%esp)       // Copy the saved CS
    movl %es:1*4(%esi), %eax
    movl %eax, 11*4(%esp)       // Copy the saved EFLAGS
    movl %cs, %ebx              // Move the current CS into EBX
    xorl %eax, %ebx             // XOR the saved & current CS values
    movl %ebx, 9*4(%esp)        // Store the result
    testl $0x3, %ebx            // If the lower 3 bits are 0, the old & new CS
    jz 1f                       // values are at the same privilege level.
    movl %es:3*4(%esi), %eax    // Otherwise, a privilege level switch occured.
    movl %eax, 13*4(%esp)       // Copy the saved ESP
    movl %es:4*4(%esi), %eax    
    movl %eax, 14*4(%esp)       // Copy the saved SS
    jmp 2f                      
1:  movl orig_far_sp, %eax      // Load the original ESP into EAX
    addl $(3*4), %eax           // Clear the exception frame off the saved stack
    movl %eax, 13*4(%esp)       // Copy the result into the new exp. frame
    movl orig_far_sp + 4, %eax  
    movl %eax, 14*4(%esp)       // Copy the saved SS
2:
.endm

.macro stackexit
    mov %esp, %edi          // Copy the stack pointer into EDI
    addl $(15*4), %esp      // Clear the exception frame off the stack
    movl 13*4(%edi), %eax   // Copy the exp. frame ESP into EAX
    movl 14*4(%edi), %ecx   // Copy the exp. frame SS into ECX
    testl $0x3, %ebx        // EBX still contains OLD_CS ^ NEW_CS, so a zero
    jz 1f                   // here indicates no privilege change
    lss orig_far_sp, %esp   // Otherwise, load the original SP into SS:ESP
    movl %eax, 3*4(%esp)    // Copy EAX into the original exception frame's ESP
    movl %ecx, 4*4(%esp)    // Copy ECX into the original exception frame's SS
    jmp 2f                  
1:  movl %eax, %esp         // Load the exp. frame ESP into ESP
    movl %ecx, %ss          // Load the exp. frame SS into SS
    subl $(3*4), %esp       // Make space for an exception frame on the new stack
    movl 10*4(%edi), %eax   
    movl %eax, (%esp)       // Copy the exp. frame EIP into the new exp. frame
    movl 11*4(%edi), %eax
    movl %eax, 1*4(%esp)    // Copy the exp. frame CS into the new exp. frame
    movl 12*4(%edi), %eax
    movl %eax, 2*4(%esp)    // Copy the exp. frame EFLAGS into the new exp. frame
2:  movl (%edi), %ebx       // Restore Registers
    movl 1*4(%edi), %ecx    
    movl 2*4(%edi), %edx
	movl 5*4(%edi), %ebp
	movl 6*4(%edi), %eax
	mov  7*4(%edi), %ds
	mov  8*4(%edi), %es
	movb $0,1(%esi)         // ESI contains a pointer to the cpu_info struct;
                            // this line reenables event delivery.
	movl 3*4(%edi), %esi
	movl 1*4(%edi), %edi
    iret
.endm

.macro zeroentry sym
    pushl $0            /* Push a fake error code */
    pushl $\sym         /* Push the trap handler address */
    jmp do_exception
.endm
    
.macro errorentry sym
    pushl $\sym
    jmp do_exception
.endm

.text

do_exception:
    pushl %ds
	pushl %eax
	xorl %eax, %eax
	pushl %ebp
	pushl %edi
	pushl %esi
	pushl %edx
	decl %eax			# eax = -1
	pushl %ecx
	pushl %ebx
	cld
	movl %es, %ecx
	movl ES(%esp), %edi		# get the function address
	movl ORIG_EAX(%esp), %edx	# get the error code
	movl %eax, ORIG_EAX(%esp)
	movl %ecx, ES(%esp)
	movl $(FLAT_KERNEL_DS), %ecx
	movl %ecx, %ds
	movl %ecx, %es
	movl %esp,%eax			# pt_regs pointer
    pushl %edx
    pushl %eax
	call *%edi
    movb CS(%esp),%cl
    addl $8,%esp
    RESTORE_ALL
    iret

/** 
 * A note on the "critical region" in our callback handler.
 * We want to avoid stacking callback handlers due to events occurring
 * during handling of the last event. To do this, we keep events disabled
 * until weve done all processing. HOWEVER, we must enable events before
 * popping the stack frame (cant be done atomically) and so it would still
 * be possible to get enough handler activations to overflow the stack.
 * Although unlikely, bugs of that kind are hard to track down, so we'd
 * like to avoid the possibility.
 * So, on entry to the handler we detect whether we interrupted an
 * existing activation in its critical region -- if so, we pop the current
 * activation and restart the handler using the previous one.
 */
EXPORT_SYMBOL(hypervisor_callback)
        stackenter
        movl EIP(%esp),%eax
        cmpl $scrit,%eax
        jb   11f
        cmpl $ecrit,%eax
        jb   critical_region_fixup
11:     push %esp
        xorl %ebp,%ebp
        call do_hypervisor_callback
        add  $4,%esp
        leal shared_info,%esi
        xorl %eax,%eax
        movb CS(%esp),%cl
    	test $2,%cl          # slow return to ring 2 or 3
        //jne  safesti
//safesti:movb $0,1(%esi)     # reenable event callbacks
scrit:  /**** START OF CRITICAL REGION ****/
        testb $0xFF,(%esi)
        jnz  14f              # process more events if necessary...
done:
        stackexit
14:     movb $1,1(%esi)
        jmp  11b
ecrit:  /**** END OF CRITICAL REGION ****/

/*
 * [How we do the fixup]. We want to merge the current stack frame with the
 * just-interrupted frame. How we do this depends on where in the critical
 * region the interrupted handler was executing, and so how many saved
 * registers are in each frame. We do this quickly using the lookup table
 * 'critical_fixup_table'. For each byte offset in the critical region, it
 * provides the number of bytes which have already been popped from the
 * interrupted stack frame. 
 */
critical_region_fixup:
        addl $critical_fixup_table-scrit,%eax
        movzbl (%eax),%eax    # %eax contains num bytes popped
        mov  %esp,%esi
        add  %eax,%esi        # %esi points at end of src region
        mov  %esp,%edi
        add  $0x34,%edi       # %edi points at end of dst region
        mov  %eax,%ecx
        shr  $2,%ecx          # convert words to bytes
        je   16f              # skip loop if nothing to copy
15:     subl $4,%esi          # pre-decrementing copy loop
        subl $4,%edi
        movl (%esi),%eax
        movl %eax,(%edi)
        loop 15b
16:     movl %edi,%esp        # final %edi is top of merged stack
        jmp  11b
         
critical_fixup_table:        
        .byte 0x00,0x00,0x00                  # testb $0xff,(%esi)
        .byte 0x00,0x00                       # jne  14f
        .byte 0x00                            # pop  %ebx
        .byte 0x04                            # pop  %ecx
        .byte 0x08                            # pop  %edx
        .byte 0x0c                            # pop  %esi
        .byte 0x10                            # pop  %edi
        .byte 0x14                            # pop  %ebp
        .byte 0x18                            # pop  %eax
        .byte 0x1c                            # pop  %ds
        .byte 0x20                            # pop  %es
        .byte 0x24,0x24,0x24                  # add  $4,%esp
        .byte 0x28                            # iret
        .byte 0x00,0x00,0x00,0x00             # movb $1,1(%esi)
        .byte 0x00,0x00                       # jmp  11b
       
/* Hypervisor uses this for application faults while it executes. */
EXPORT_SYMBOL(failsafe_callback)
      pop  %ds
      pop  %es
      pop  %fs
      pop  %gs
      iret
      
EXPORT_SYMBOL(coprocessor_error)
	zeroentry do_coprocessor_error

EXPORT_SYMBOL(simd_coprocessor_error)
	zeroentry do_simd_coprocessor_error

EXPORT_SYMBOL(device_not_available)
    iret

EXPORT_SYMBOL(debug)
	zeroentry do_debug

EXPORT_SYMBOL(int3)
	zeroentry do_int3

EXPORT_SYMBOL(overflow)
	zeroentry do_overflow

EXPORT_SYMBOL(bounds)
	zeroentry do_bounds

EXPORT_SYMBOL(invalid_op)
	zeroentry do_invalid_op

EXPORT_SYMBOL(coprocessor_segment_overrun)
	zeroentry do_coprocessor_segment_overrun

EXPORT_SYMBOL(invalid_TSS)
	errorentry do_invalid_TSS

EXPORT_SYMBOL(segment_not_present)
	errorentry do_segment_not_present

EXPORT_SYMBOL(stack_segment)
	errorentry do_stack_segment

EXPORT_SYMBOL(general_protection)
	errorentry do_general_protection

EXPORT_SYMBOL(alignment_check)
	errorentry do_alignment_check

EXPORT_SYMBOL(divide_error)
	zeroentry do_divide_error
    
EXPORT_SYMBOL(machine_check)
	zeroentry do_machine_check

EXPORT_SYMBOL(spurious_interrupt_bug)
	zeroentry do_spurious_interrupt_bug
	
EXPORT_SYMBOL(page_fault)
    errorentry do_page_fault

