/******************************************************************************
 * events.c
 * 
 * Event handling
 * 
 * Copyright (c) 2010-2011, Pat Gavlin
 * Copyright (c) 2002-2003, K A Fraser
 * Copyright (c) 2005, Grzegorz Milos, gm281@cam.ac.uk,Intel Research Cambridge
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

#include <stdint.h>
#include "events.h"
#include "hypercall.h"
#include "util.h"

// Number of event channels
#define N_EVTCHN 1024

// Event handler type
struct evtchn_action {
	evtchn_handler_t handler;
	void *data;
};
typedef struct evtchn_action evtchn_action_t;

// Event handlers array & bound ports bitmap
static evtchn_action_t evtchn_actions[N_EVTCHN];
static unsigned long bound_ports[N_EVTCHN/(8*sizeof(unsigned long))];

// Shared info
extern shared_info_t shared_info;

// Generic event handler
void do_event(evtchn_port_t port, regs_t * regs);

// Active event channel macro
#define active_evtchns(sh,idx) \
	((sh).evtchn_pending[idx] & \
	 ~(sh).evtchn_mask[idx])

// Two-level event system:
// - evtchn_pending_sel (unsigned long): contains one bit per 
//   (sizeof(unsigned long) * 4) channels
// - evtchn_pending and evtchn_mask: arrays containing one 
//   (sizeof(unsigned long) * 4)-bit entry for each bit of evtchn_pending_sel
void
do_hypervisor_callback(regs_t *regs) {
	unsigned long l1_selector, l2_selector, l1_index, l2_index;
	unsigned int port;
	vcpu_info_t *vcpu_info = &shared_info.vcpu_info[0];
 
	// Clear the event pending flag
	vcpu_info->evtchn_upcall_pending = 0;

	// clear the evtchn selector (atomically?)
	l1_selector = xchg(&vcpu_info->evtchn_pending_sel, 0);

	// Process until all pending events have been handled
	while (l1_selector != 0) {
		l1_index = __ffs(l1_selector);		// Find first bit set
		l1_selector &= ~(1UL << l1_index);	// Mask off this bit
		
		// Demux the evtchn selector & process until all of its
		// pending events have been handled
		while ( (l2_selector = active_evtchns(shared_info, l1_index)) != 0 ) {
			l2_index = __ffs(l2_selector);	  // Find first set bit
			l2_selector &= ~(1UL << l2_index);  // Mask off this bit

			// Compute the port # of the event and handle it
			port = (l1_index * (sizeof(unsigned long) * 8)) + l2_index;
			do_event(port, regs);
		}
	}
}

void
force_evtchn_callback(void) {
	int save;
	vcpu_info_t *vcpu;
	vcpu = &shared_info.vcpu_info[0];
	save = vcpu->evtchn_upcall_mask;

	while (vcpu->evtchn_upcall_pending) {
		vcpu->evtchn_upcall_mask = 1;
		barrier();
		do_hypervisor_callback(0);
		barrier();
		vcpu->evtchn_upcall_mask = save;
		barrier();
	};
}

void
mask_evtchn(uint32_t port) {
	synch_set_bit(port, &shared_info.evtchn_mask[0]);
}

void
unmask_evtchn(uint32_t port) {
	vcpu_info_t *vcpu_info = &shared_info.vcpu_info[0];

	synch_clear_bit(port, &shared_info.evtchn_mask[0]);

	/*
	 * The following is basically the equivalent of 'hw_resend_irq'. Just like
	 * a real IO-APIC we 'lose the interrupt edge' if the channel is masked.
	 */
	if (synch_test_bit(port, &shared_info.evtchn_pending[0]) && 
		 !synch_test_and_set_bit(port / (sizeof(unsigned long) * 8),
			  &vcpu_info->evtchn_pending_sel) )
	{
		vcpu_info->evtchn_upcall_pending = 1;
		if ( !vcpu_info->evtchn_upcall_mask )
			force_evtchn_callback();
	}
}

static inline void
clear_evtchn(uint32_t port) {
	synch_clear_bit(port, &shared_info.evtchn_pending[0]);
}

void
do_event(evtchn_port_t port, regs_t *regs) {
	evtchn_action_t *action;

	synch_clear_bit(port, &shared_info.evtchn_pending[0]);
	//clear_evtchn(port);

	if (port >= N_EVTCHN) return;

	action = &evtchn_actions[port];
	action->handler(port, regs, action->data);

	return;
}

void
default_handler(evtchn_port_t port, regs_t *regs, void *data) {
}

evtchn_port_t
alloc_unbound_evtchn(evtchn_port_t port) {
}

evtchn_port_t
bind_virq(uint32_t virq) {
	evtchn_bind_virq_t op = {virq, 0, 0};
	HYPERVISOR_event_channel_op(EVTCHNOP_bind_virq, &op);
	return op.port;
}

evtchn_port_t
bind_evtchn(evtchn_port_t port, evtchn_handler_t handler, void *data) {
	evtchn_actions[port].data = data;
	wmb();
	evtchn_actions[port].handler = handler;
	set_bit(port, bound_ports);

	return port;
}

void
unbind_evtchn(evtchn_port_t port) {
	struct evtchn_close close;
	int rc;

	synch_set_bit(port, &shared_info.evtchn_mask[0]);
	synch_clear_bit(port, &shared_info.evtchn_pending[0]);

	evtchn_actions[port].handler = default_handler;
	wmb();
	evtchn_actions[port].data = 0;
	clear_bit(port, bound_ports);

	close.port = port;
	rc = HYPERVISOR_event_channel_op(EVTCHNOP_close, &close);	  
}

char irqstack[16384];

#if defined(__x86_32__)

struct far_pointer {
	unsigned long offset;
	unsigned short seg;
};
typedef struct far_pointer far_pointer_t;

far_pointer_t exp_far_sp;
far_pointer_t orig_far_sp;

#elif defined(__x86_64__)

static struct pda {
	int irqcount;	   /* offset 0 (used in x86_64.S) */
	char *irqstackptr;  /*		8 */
} cpu0_pda;
#endif

/*
 * Initially all events are without a handler and disabled
 */
void
event_init(void) {
	int i;
	
#if defined(__x86_32__)
	exp_far_sp.seg = (unsigned short)(FLAT_KERNEL_SS);
	exp_far_sp.offset = (((unsigned long)irqstack + 16384) & ~(8191));
#elif defined(__x86_64__)
	__asm__ __volatile__ ("movl %0,%%fs ; movl %0,%%gs" :: "r" (0));
	wrmsrl(0xc0000101, &cpu0_pda); /* 0xc0000101 is MSR_GS_BASE */
	cpu0_pda.irqcount = -1;
	cpu0_pda.irqstackptr = (void*) (((unsigned long)irqstack + 16384) & ~(8191));
#endif

	/* initialize event handler */
	for ( i = 0; i < N_EVTCHN; i++ ) {
		evtchn_actions[i].handler = default_handler;
		synch_set_bit(i, &shared_info.evtchn_mask[0]);
	}
}
