#include <stdint.h>
#include <xen/xen.h>
#include <xen/event_channel.h>
#include <xen/sched.h>
#include <xen/vcpu.h>
#include "hypercall.h"
#include "util.h"

extern shared_info_t shared_info;

// Safe copy of the Xen time info
static vcpu_time_info_t time_info;

// Timer event channel
evtchn_port_t timer_evtchn;

int64_t getsystemtime();

/**
 * native_read_tsc: Reads a value from the TSC
 *
 * Returns: The current TSC value
 */
uint64_t
native_read_tsc() {
	uint64_t tsc;
	__asm__ __volatile__(
		"rdtsc\n" \
		"movl %%eax, (%0)\n" \
		"movl %%edx, 4(%0)\n" : : "r" (&tsc) : "eax", "edx", "memory"
	);
	return tsc;
}

/**
 * scale_delta: Scales and shifts a time difference 
 *
 * delta:	The difference to scale
 * mul_frac: The scale factor
 * shift:	The number of bits to shift 
 *
 * Returns: The scaled & shifted result
 */
static inline uint64_t
scale_delta(uint64_t delta, uint32_t mul_frac, int shift) {
		uint64_t product;
#ifdef __i386__
		uint32_t tmp1, tmp2;
#endif

		if (shift < 0)
				delta >>= -shift;
		else
				delta <<= shift;

#ifdef __i386__
		__asm__ (
				"mul  %5	   ; "
				"mov  %4,%%eax ; "
				"mov  %%edx,%4 ; "
				"mul  %5	   ; "
				"xor  %5,%5	; "
				"add  %4,%%eax ; "
				"adc  %5,%%edx ; "
				: "=A" (product), "=r" (tmp1), "=r" (tmp2)
				: "a" ((uint32_t)delta), "1" ((uint32_t)(delta >> 32)), "2" (mul_frac) );
#elif defined(__x86_64__)
		__asm__ (
				"mul %%rdx ; shrd $32,%%rdx,%%rax"
				: "=a" (product) : "0" (delta), "d" ((uint64_t)mul_frac) );
#else
#error Architecture unsupported.
#endif

		return product;
}

/**
 * get_nsec_offset: Gets the nanosecond offset from a time_info struct
 *
 * ti: Pointer to the time info struct to use
 *
 * Returns: The computed nanosecond offset
 */
static uint64_t
get_nsec_offset(vcpu_time_info_t *ti) {
		uint64_t delta = native_read_tsc() - ti->tsc_timestamp;
		return scale_delta(delta, ti->tsc_to_system_mul, ti->tsc_shift);
}

/**
 * getsystemtime: Get the current system time
 *
 * Returns: The number of nanoseconds since boot time
 */
int64_t
getsystemtime() {
	vcpu_time_info_t *ti = &(shared_info.vcpu_info[0].time);
	
	// Wait for our version of the time struct to match the current version
	if (ti->version != time_info.version) {
		while ((ti->version & 1) != 0) 
			rmb();
		memcpy(&time_info, ti, sizeof(time_info));
	}
	
	// Compute and return the time
	return time_info.system_time + get_nsec_offset(&time_info);
}

/**
 * time_init: Initialize the time/timer system
 * 
 * NOTE: This function is incomplete!  You will need 
 *       to add code here in order to use timer events.
 *
 * Returns: 0 on success, -1 otherwise
 */
int
time_init() {
	// Stop the periodic timer
	if (HYPERVISOR_vcpu_op(VCPUOP_stop_periodic_timer, 0, 0) != 0) {
		return -1;
	}

	// Copy the time info struct somewhere safe
	memcpy(&time_info, &(shared_info.vcpu_info[0].time), sizeof(time_info));

	return 0;
}


