#include <stdint.h>
#include <xen/xen.h>
#include "hypercall.h"
#include "printf.h"
#include "traps.h"

extern shared_info_t shared_info;

/**
 * These are assembler stubs in entry.<arch>.S.
 * They are the actual entry points for virtual exceptions.
 */
void divide_error(void);
void debug(void);
void int3(void);
void overflow(void);
void bounds(void);
void invalid_op(void);
void device_not_available(void);
void coprocessor_segment_overrun(void);
void invalid_TSS(void);
void segment_not_present(void);
void stack_segment(void);
void general_protection(void);
void page_fault(void);
void coprocessor_error(void);
void simd_coprocessor_error(void);
void alignment_check(void);
void spurious_interrupt_bug(void);
void machine_check(void);

/**
 * dump_regs: Dumps pre-exception register contents to the console
 *
 * NOTE: It is recommended that you add code here in accordance
 *       with the above description.
 * 
 * regs: Register contents to dump
 */
void
dump_regs(regs_t *regs) {
}

/**
 * do_crash: Crashes the domain
 */
void
do_crash(regs_t *regs) {
	if (regs != NULL)
		dump_regs(regs);

	sched_shutdown_t sd = {SHUTDOWN_crash};
	HYPERVISOR_sched_op(SCHEDOP_shutdown, &sd);
}

/**
 * do_trap: Generic trap handler
 *
 * trapnr:	 Trap number
 * str:		Error string
 * regs:	   Pre-exception register contents
 * error_code: Error code, if any
 */
static void
do_trap(int trapnr, char *str, regs_t *regs, unsigned long error_code) {
	do_crash(regs);
}

/**
 * DO_ERROR, DO_ERROR_INFO: Convenience wrappers around do_trap
 */
#define DO_ERROR(trapnr, str, name) \
void do_##name(regs_t *regs, unsigned long error_code) { \
	do_trap(trapnr, str, regs, error_code); \
}

#define DO_ERROR_INFO(trapnr, str, name, sicode, siaddr) \
void do_##name(regs_t *regs, unsigned long error_code) { \
	do_trap(trapnr, str, regs, error_code); \
}

DO_ERROR_INFO( 0, "divide error", divide_error, FPE_INTDIV, regs->eip)
DO_ERROR( 3, "int3", int3)
DO_ERROR( 4, "overflow", overflow)
DO_ERROR( 5, "bounds", bounds)
DO_ERROR_INFO( 6, "invalid operand", invalid_op, ILL_ILLOPN, regs->eip)
DO_ERROR( 7, "device not available", device_not_available)
DO_ERROR( 9, "coprocessor segment overrun", coprocessor_segment_overrun)
DO_ERROR(10, "invalid TSS", invalid_TSS)
DO_ERROR(11, "segment not present", segment_not_present)
DO_ERROR(12, "stack segment", stack_segment)
DO_ERROR_INFO(17, "alignment check", alignment_check, BUS_ADRALN, 0)
DO_ERROR(18, "machine check", machine_check)

/**
 * do_<exception>: Handlers for specific exceptions 
 */
void
do_page_fault(regs_t *regs, unsigned long error_code) {
	do_crash(regs);
}

void do_general_protection(regs_t *regs, long error_code) {
	do_crash(regs);
}

void
do_debug(regs_t *regs) {
	do_crash(regs);
}

void
do_coprocessor_error(regs_t *regs) {
	do_crash(regs);
}

void
simd_math_error(void *eip) {
	do_crash(NULL);
}

void
do_simd_coprocessor_error(regs_t *regs) {
	do_crash(regs);
}

void
do_spurious_interrupt_bug(regs_t *regs) {
	do_crash(regs);
}

/**
 * Submit a virtual IDT to the hypervisor. This consists of tuples
 * (interrupt vector, privilege ring, CS:EIP of handler).
 * The 'privilege ring' field specifies the least-privileged ring that
 * can trap to that vector using a software-interrupt instruction (INT).
 */
static trap_info_t trap_table[] = {
	{  0, 0, FLAT_KERNEL_CS, (unsigned long)divide_error				},
	{  1, 0, FLAT_KERNEL_CS, (unsigned long)debug					   },
	{  3, 3, FLAT_KERNEL_CS, (unsigned long)int3						},
	{  4, 3, FLAT_KERNEL_CS, (unsigned long)overflow					},
	{  5, 3, FLAT_KERNEL_CS, (unsigned long)bounds					  },
	{  6, 0, FLAT_KERNEL_CS, (unsigned long)invalid_op				  },
	{  7, 0, FLAT_KERNEL_CS, (unsigned long)device_not_available		},
	{  9, 0, FLAT_KERNEL_CS, (unsigned long)coprocessor_segment_overrun },
	{ 10, 0, FLAT_KERNEL_CS, (unsigned long)invalid_TSS				 },
	{ 11, 0, FLAT_KERNEL_CS, (unsigned long)segment_not_present		 },
	{ 12, 0, FLAT_KERNEL_CS, (unsigned long)stack_segment			   },
	{ 13, 0, FLAT_KERNEL_CS, (unsigned long)general_protection		  },
	{ 14, 0, FLAT_KERNEL_CS, (unsigned long)page_fault				  },
	{ 15, 0, FLAT_KERNEL_CS, (unsigned long)spurious_interrupt_bug	  },
	{ 16, 0, FLAT_KERNEL_CS, (unsigned long)coprocessor_error		   },
	{ 17, 0, FLAT_KERNEL_CS, (unsigned long)alignment_check			 },
	{ 19, 0, FLAT_KERNEL_CS, (unsigned long)simd_coprocessor_error	  },
	{  0, 0, 0, 0 }
};

/**
 * trap_init: Initializes the trap subsystem
 */
void
trap_init() {
	// Pass the virtual IDT to the hypervisor
	HYPERVISOR_set_trap_table(trap_table);
}
